#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned int uint;

/* ------ Compression ------ */

/*
 * Mode 2 Row compression routine for the HP DeskJet & LaserJet IIp.
 * Compresses data from row up to end_row, storing the result
 * starting at compressed.  Returns the number of bytes stored.
 * Runs of K<=127 literal bytes are encoded as K-1 followed by
 * the bytes; runs of 2<=K<=127 identical bytes are encoded as
 * 257-K followed by the byte.
 * In the worst case, the result is N+(N/127)+1 bytes long,
 * where N is the original byte count (end_row - row).
 * To speed up the search, we examine an entire word at a time.
 * We will miss a few blocks of identical bytes; tant pis.
 */
 
int mode2compress(byte *row, byte *end_row, byte *compressed)
{    
    register byte *exam; /* word being examined in the row to compress */
    register byte *cptr = compressed; /* output pointer into compressed bytes */
    int i, count, len;
    byte test;

    exam = row;
    while (1)
    {                        
        test = *exam++;
        /* Advance exam until test==*exam  or exam==end_row */
        while ((test != *exam) && (exam < end_row))
            test = *exam++;                    
        /* row points to start of differing bytes,
           exam points to start of consequtive series 
                    or to end of row */
        if (exam<end_row) exam--;
        len=exam-row;   
        while (len>0)
        {
            count=len;
            if (count>127) count=127;
            *cptr++=count-1;  
            for (i=0;i<count;i++) *cptr++ = *row++;
            len-=count;
        }     
        if (exam>=end_row) break;     /* done */             
        exam++;     /* skip first same byte */
        while ((test == *exam) && (exam < end_row))  /* skip all same bytes */
            exam++;
        /* exam points now first different word or to end of data */
        len = exam-row;
        while (len>0)
        {     
            count=len;
            if (count>127) count=127;
            *cptr++=(257-count);
            *cptr++=test;             
            len-=count;
        }                
        if (exam>=end_row) break;            /* end of data */
        row = exam;    /* row points to first dissimular byte */
    }                                          
    return (cptr-compressed);
}

/*
 * Mode 3 compression routine for the HP LaserJet III family.
 * Compresses bytecount bytes starting at current, storing the result
 * in compressed, comparing against and updating previous.
 * Returns the number of bytes stored.  In the worst case,
 * the number of bytes is bytecount+(bytecount/8)+1.
 */

int mode3compress (int bytecount, char *current, char *previous, char *compressed)
{
    char *cur = current;
    char *prev = previous;
    char *out = compressed;
    char *end = current + bytecount;

    while (cur < end)
    {				/* Detect a maximum run of unchanged bytes. */
	char *run = cur;
	char *diff;
	char *stop;
	int offset, cbyte;

	while (cur < end && *cur == *prev)
	{
	    cur++, prev++;
	}
	if (cur == end)
	    break;		/* rest of row is unchanged */
	/* Detect a run of up to 8 changed bytes. */
	/* We know that *cur != *prev. */
	diff = cur;
	stop = (end - cur > 8 ? cur + 8 : end);
	do
	{
	    *prev++ = *cur++;
	}
	while (cur < stop && *cur != *prev);
	/* Now [run..diff) are unchanged, and */
	/* [diff..cur) are changed. */
	/* Generate the command byte(s). */
	offset = diff - run;
	cbyte = (cur - diff - 1) << 5;
	if (offset < 31)
	    *out++ = cbyte + offset;
	else
	{
	    *out++ = cbyte + 31;
	    offset -= 31;
	    while (offset >= 255)
		*out++ = 255, offset -= 255;
	    *out++ = offset;
	}
	/* Copy the changed data. */
	while (diff < cur)
	    *out++ = *diff++;
    }
    return out - compressed;
}

int mode9compress(int bytecount, const byte *current, byte *previous, byte *compressed)
{
  const byte *cur = current;
  byte *prev = previous;
  byte *out = compressed;
  const byte *end = current + bytecount;

  while ( cur < end )
    {	/* Detect a run of unchanged bytes. */
      const byte *run = cur;
      const byte *diff;
      int offset;
      while ( cur < end && *cur == *prev )
	{	cur++, prev++;
	}
      if ( cur == end ) break;	/* rest of row is unchanged */
      /* Detect a run of changed bytes. */
      /* We know that *cur != *prev. */
      diff = cur;
      do
	{	*prev++ = *cur++;
	}
      while ( cur < end && *cur != *prev );
      /* Now [run..diff) are unchanged, and */
      /* [diff..cur) are changed. */
      offset = diff - run;
      {
	const byte  *stop_test = cur - 4;
	int   dissimilar, similar;
	
	while (diff < cur) {
	  const byte *compr = diff;
	  const byte *next;	/* end of run */
	  byte value = 0;
	  while (diff <= stop_test &&
		 ((value = *diff) != diff[1] ||
		  value != diff[2] ||
		  value != diff[3]))
	    diff++;
	  
	  /* Find out how long the run is */
	  if (diff > stop_test)	/* no run */
	    next = diff = cur;
	  else {
	    next = diff + 4;
	    while (next < cur && *next == value)
	      next++;
	  }
	  
#define MAXOFFSETU 15
#define MAXCOUNTU 7
	  /* output 'dissimilar' bytes, uncompressed */
	  if ((dissimilar = diff - compr))
	    {
	      int temp, i;

	      if ((temp = --dissimilar) > MAXCOUNTU) temp = MAXCOUNTU;
	      if (offset < MAXOFFSETU)
		*out++ = (offset << 3) | (byte)temp;
	      else {
		*out++ = (MAXOFFSETU << 3) | (byte)temp;
		offset -= MAXOFFSETU;
		while (offset >= 255) {
		  *out++ = 255;
		  offset -= 255;
		}
		*out++ = offset;
	      }
	      if (temp == MAXCOUNTU) {
		temp = dissimilar - MAXCOUNTU;
		while (temp >= 255) {
		  *out++ = 255;
		  temp -= 255;
		}
		*out++ = (byte)temp;
	      }
	      for (i = 0; i <= dissimilar; i++) *out++ = *compr++;
	      offset = 0;
	    } /* end uncompressed */
	  
#define MAXOFFSETC 3
#define MAXCOUNTC 31
	  /* output 'similar' bytes, run-length endcoded */
	  if ((similar = next - diff))
	    {
	      int temp;

	      if ((temp = (similar -= 2)) > MAXCOUNTC) temp = MAXCOUNTC;
	      if (offset < MAXOFFSETC)
		*out++ = 0x80 | (offset << 5) | (byte)temp;
	      else {
		*out++ = 0x80 | (MAXOFFSETC << 5) | (byte)temp;
		offset -= MAXOFFSETC;
		while (offset >= 255) {
		  *out++ = 255;
		  offset -= 255;
		}
		*out++ = offset;
	      }
	      if (temp == MAXCOUNTC) {
		temp = similar - MAXCOUNTC;
		while (temp >= 255) {
		  *out++ = 255;
		  temp -= 255;
		}
		*out++ = (byte)temp;
	      }
	      *out++ = value;
	      offset = 0;
	    } /* end compressed */
	  
	  diff = next;
	}
      }
    }
  return out - compressed;
}

